@@ -13,7 +13,10 @@ import io
 import logging
 import os
 import sys
-from typing import Any, Dict, Optional, Union
+from pathlib import Path
+from typing import Any, Dict, Optional, TextIO, Union
+
+from cobib.utils.rel_path import RelPath
 
 LOGGER = logging.getLogger(__name__)
 
@@ -29,7 +32,7 @@ ANSI_COLORS = [
 ]
 
 
-class Config(dict):
+class Config(Dict[str, Any]):
     """coBib's configuration class.
 
     This class wraps the `dict` type and exposes the dictionary keys as attributes to ease access
@@ -48,9 +51,9 @@ class Config(dict):
     """The *legacy* XDG-based standard configuration location.
     This will be removed on January 1st, 2020."""
 
-    DEFAULTS: Dict = {
+    DEFAULTS: Dict[str, Any] = {
         "logging": {
-            "logfile": os.path.expanduser("~/.cache/cobib/cobib.log"),
+            "logfile": "~/.cache/cobib/cobib.log",
         },
         "commands": {
             "edit": {
@@ -60,15 +63,25 @@ class Config(dict):
             "open": {
                 "command": "xdg-open" if sys.platform.lower() == "linux" else "open",
             },
-            "search": {"grep": "grep", "ignore_case": False},
+            "search": {
+                "grep": "grep",
+                "grep_args": [],
+                "ignore_case": False,
+            },
         },
         "database": {
-            "file": os.path.expanduser("~/.local/share/cobib/literature.yaml"),
+            "file": "~/.local/share/cobib/literature.yaml",
             "format": {
-                "month": int,
                 "suppress_latex_warnings": True,
             },
             "git": False,
+            "stringify": {
+                "list_separator": {
+                    "file": ", ",
+                    "tags": ", ",
+                    "url": ", ",
+                },
+            },
         },
         "parsers": {
             "bibtex": {
@@ -120,11 +133,18 @@ class Config(dict):
                 "show": "ENTER",
             },
         },
+        "utils": {
+            "file_downloader": {
+                "default_location": "~/.local/share/cobib",
+                "url_map": {},
+            },
+            "journal_abbreviations": [],
+        },
     }
     """The default settings."""
 
     # pylint: disable=super-init-not-called
-    def __init__(self, value: Optional[Dict] = None) -> None:
+    def __init__(self, value: Optional[Dict[str, Any]] = None) -> None:
         """Initializer of the recursive, attribute-access, dict-like configuration object.
 
         The initializer does nothing when `None` is given.
@@ -168,6 +188,10 @@ class Config(dict):
     MARKER = object()
     """A helper object for detecting the nested recursion-threshold."""
 
+    EXCEPTIONAL_KEYS = {"url_map"}
+    """A set of exceptional keys which do not cause automatic item creation. This is required in
+    order to support `dict`-like configuration options properly."""
+
     def __getitem__(self, key: str) -> Any:
         """Gets a key from the configuration object's dictionary.
 
@@ -181,7 +205,7 @@ class Config(dict):
             The value of the queried attribute.
         """
         found = self.get(key, Config.MARKER)
-        if found is Config.MARKER:
+        if found is Config.MARKER and key not in Config.EXCEPTIONAL_KEYS:
             found = Config()
             super().__setitem__(key, found)
         return found
@@ -213,7 +237,7 @@ class Config(dict):
             self.__setitem__(key, copy.deepcopy(value))
 
     @staticmethod
-    def load(configpath: Union[str, io.TextIOWrapper] = None) -> None:
+    def load(configpath: Optional[Union[str, Path, TextIO, io.TextIOWrapper]] = None) -> None:
         """Loads another configuration object at runtime.
 
         WARNING: The new Python-like configuration allows essentially arbitrary Python code so it is
@@ -223,25 +247,17 @@ class Config(dict):
             configpath: the path to the configuration.
         """
         if configpath is not None:
-            if isinstance(configpath, io.TextIOWrapper):
+            if isinstance(configpath, (TextIO, io.TextIOWrapper)):
                 configpath = configpath.name
-            LOGGER.info("Loading configuration from %s", configpath)
-        elif os.path.exists(os.path.expanduser(Config.XDG_CONFIG_FILE)):
-            LOGGER.info(
-                "Loading configuration from default location: %s",
-                os.path.expanduser(Config.XDG_CONFIG_FILE),
-            )
-            configpath = os.path.expanduser(Config.XDG_CONFIG_FILE)
-        elif os.path.exists(os.path.expanduser(Config.LEGACY_XDG_CONFIG_FILE)):
-            LOGGER.info(
-                "Loading configuration from default location: %s",
-                os.path.expanduser(Config.LEGACY_XDG_CONFIG_FILE),
-            )
-            configpath = os.path.expanduser(Config.LEGACY_XDG_CONFIG_FILE)
-        else:
-            return
+        elif RelPath(Config.XDG_CONFIG_FILE).exists():
+            configpath = RelPath(Config.XDG_CONFIG_FILE).path
+        elif RelPath(Config.LEGACY_XDG_CONFIG_FILE).exists():  # pragma: no cover
+            configpath = RelPath(Config.LEGACY_XDG_CONFIG_FILE).path  # pragma: no cover
+        else:  # pragma: no cover
+            return  # pragma: no cover
+        LOGGER.info("Loading configuration from default location: %s", configpath)
 
-        if os.path.exists(os.path.expanduser(Config.LEGACY_XDG_CONFIG_FILE)):
+        if RelPath(Config.LEGACY_XDG_CONFIG_FILE).exists():
             msg = (
                 "The configuration mechanism of coBib underwent a major re-design for version 3.0! "
                 "This means, that the old `INI`-style configuration is deprecated and will be "
@@ -255,7 +271,7 @@ class Config(dict):
             print("\x1b[1;37;41m#############\x1b[0m", file=sys.stderr)
             print("\x1b[1;37;41m## WARNING ##\x1b[0m", file=sys.stderr)
             print("\x1b[1;37;41m#############\x1b[0m", file=sys.stderr)
-            print(msg, file=sys.stderr)
+            LOGGER.warning(msg)
 
         spec = importlib.util.spec_from_file_location("config", configpath)
         if spec is None:
@@ -276,7 +292,7 @@ class Config(dict):
             sys.exit(1)
 
     @staticmethod
-    def load_legacy_config(configpath: Union[str, io.TextIOWrapper]) -> None:
+    def load_legacy_config(configpath: Union[str, Path, TextIO, io.TextIOWrapper]) -> None:
         # pylint: disable=too-many-branches,too-many-nested-blocks
         """Loads a legacy `INI`-style configuration file.
 
@@ -291,6 +307,16 @@ class Config(dict):
         ini_conf.optionxform = str  # type: ignore
         ini_conf.read(configpath)
 
+        def _ignore_option(section: str, field: str, value: Optional[Any] = None) -> None:
+            if value is not None:  # pragma: no cover
+                LOGGER.warning(  # pragma: no cover
+                    "Ignoring unknown option for %s/%s = %s", section, field, value
+                )
+            else:  # pragma: no cover
+                LOGGER.warning(  # pragma: no cover
+                    "Ignoring unknown setting %s", f"{section}/{field}"
+                )
+
         # We need to manually iterate all sections and fields because some settings need to be moved
         # and/or need to be converted to the correct Python types
         for section in ini_conf.sections():
@@ -301,11 +327,9 @@ class Config(dict):
                     elif field in ["git"]:
                         try:
                             config.database[field] = ini_conf[section].getboolean(field)
-                        except ValueError as exc:
-                            LOGGER.error(exc)
-                            LOGGER.warning(
-                                "Ignoring unknown option for %s/%s = %s", section, field, value
-                            )
+                        except ValueError as exc:  # pragma: no cover
+                            LOGGER.error(exc)  # pragma: no cover
+                            _ignore_option(section, field, value)  # pragma: no cover
                     elif field == "open":
                         config.commands.open.command = value
                     elif field == "grep":
@@ -313,7 +337,7 @@ class Config(dict):
                     elif field == "search_ignore_case":
                         config.commands.search.ignore_case = ini_conf[section].getboolean(field)
                     else:
-                        LOGGER.warning("Ignoring unknown setting %s", f"{section}/{field}")
+                        _ignore_option(section, field)  # pragma: no cover
             elif section == "FORMAT":
                 for field, value in dict(ini_conf[section]).items():
                     if field == "default_entry_type":
@@ -321,22 +345,18 @@ class Config(dict):
                     elif field == "ignore_non_standard_types":
                         try:
                             config.parsers.bibtex[field] = ini_conf[section].getboolean(field)
-                        except ValueError as exc:
-                            LOGGER.error(exc)
-                            LOGGER.warning(
-                                "Ignoring unknown option for %s/%s = %s", section, field, value
-                            )
+                        except ValueError as exc:  # pragma: no cover
+                            LOGGER.error(exc)  # pragma: no cover
+                            _ignore_option(section, field, value)  # pragma: no cover
                     elif field == "month":
                         if value == "int":
                             config.database.format.month = int
                         elif value == "str":
                             config.database.format.month = str
                         else:
-                            LOGGER.warning(
-                                "Ignoring unknown option for %s/%s = %s", section, field, value
-                            )
+                            _ignore_option(section, field, value)  # pragma: no cover
                     else:
-                        LOGGER.warning("Ignoring unknown setting %s", f"{section}/{field}")
+                        _ignore_option(section, field)  # pragma: no cover
             elif section == "TUI":
                 for field, value in dict(ini_conf[section]).items():
                     if field == "default_list_args":
@@ -344,21 +364,17 @@ class Config(dict):
                     elif field in ["prompt_before_quit", "reverse_order"]:
                         try:
                             config.tui[field] = ini_conf[section].getboolean(field)
-                        except ValueError as exc:
-                            LOGGER.error(exc)
-                            LOGGER.warning(
-                                "Ignoring unknown option for %s/%s = %s", section, field, value
-                            )
+                        except ValueError as exc:  # pragma: no cover
+                            LOGGER.error(exc)  # pragma: no cover
+                            _ignore_option(section, field, value)  # pragma: no cover
                     elif field in ["scroll_offset"]:
                         try:
                             config.tui[field] = ini_conf[section].getint(field)
-                        except ValueError as exc:
-                            LOGGER.error(exc)
-                            LOGGER.warning(
-                                "Ignoring unknown option for %s/%s = %s", section, field, value
-                            )
+                        except ValueError as exc:  # pragma: no cover
+                            LOGGER.error(exc)  # pragma: no cover
+                            _ignore_option(section, field, value)  # pragma: no cover
                     else:
-                        LOGGER.warning("Ignoring unknown setting %s", f"{section}/{field}")
+                        _ignore_option(section, field)  # pragma: no cover
             elif section == "COLORS":
                 for field, value in dict(ini_conf[section]).items():
                     config.tui.colors[field] = value
@@ -366,7 +382,7 @@ class Config(dict):
                 for field, value in dict(ini_conf[section]).items():
                     config.tui.key_bindings[field.lower()] = value
             else:
-                LOGGER.warning("Ignoring unknown config section %s", section)
+                LOGGER.warning("Ignoring unknown config section %s", section)  # pragma: no cover
 
     def validate(self) -> None:
         """Validates the configuration at runtime.
@@ -406,6 +422,10 @@ class Config(dict):
             isinstance(self.commands.search.grep, str),
             "config.commands.search.grep should be a string.",
         )
+        self._assert(
+            isinstance(self.commands.search.grep_args, list),
+            "config.commands.search.grep_args should be a list.",
+        )
         self._assert(
             isinstance(self.commands.search.ignore_case, bool),
             "config.commands.search.ignore_case should be a boolean.",
@@ -419,14 +439,29 @@ class Config(dict):
             isinstance(self.database.git, bool), "config.database.git should be a boolean."
         )
         # DATABASE.FORMAT section
-        self._assert(
-            self.database.format.month in (int, str),
-            "config.database.format.month should be either the `int` or `str` type.",
-        )
+        if "month" in self["database"]["format"].keys():
+            LOGGER.warning(
+                "The config.database.format.month setting is deprecated as of version 3.1.0! "
+                "Instead, coBib will store the month as a three-letter code which is a common "
+                "format for which most citation styles include macros. See also "
+                "https://www.bibtex.com/f/month-field/"
+            )
         self._assert(
             isinstance(self.database.format.suppress_latex_warnings, bool),
             "config.database.format.suppress_latex_warnings should be a boolean.",
         )
+        self._assert(
+            isinstance(self.database.stringify.list_separator.file, str),
+            "config.database.stringify.list_separator.file should be a string.",
+        )
+        self._assert(
+            isinstance(self.database.stringify.list_separator.tags, str),
+            "config.database.stringify.list_separator.tags should be a string.",
+        )
+        self._assert(
+            isinstance(self.database.stringify.list_separator.url, str),
+            "config.database.stringify.list_separator.url should be a string.",
+        )
 
         # PARSER section
         self._assert(
@@ -454,6 +489,7 @@ class Config(dict):
 
         # TUI.COLORS section
         LOGGER.debug("Validating the TUI.COLORS configuration section.")
+        # pylint: disable=consider-iterating-dictionary
         for name in self.DEFAULTS["tui"]["colors"].keys():
             self._assert(
                 name in self.tui.colors.keys(), f"Missing config.tui.colors.{name} specification!"
@@ -475,6 +511,7 @@ class Config(dict):
 
         # TUI.KEY_BINDINGS section
         LOGGER.debug("Validating the TUI.KEY_BINDINGS configuration section.")
+        # pylint: disable=consider-iterating-dictionary
         for command in self.DEFAULTS["tui"]["key_bindings"].keys():
             self._assert(
                 command in self.tui.key_bindings.keys(),
@@ -485,6 +522,32 @@ class Config(dict):
                 isinstance(key, str), f"config.tui.key_bindings.{command} should be a string."
             )
 
+        # UTILS section
+        LOGGER.debug("Validating the UTILS configuration section.")
+        self._assert(
+            isinstance(self.utils.file_downloader.default_location, str),
+            "config.utils.file_downloader.default_location should be a string.",
+        )
+        self._assert(
+            isinstance(self.utils.file_downloader.url_map, dict),
+            "config.utils.file_downloader.url_map should be a dict.",
+        )
+        for pattern, repl in self.utils.file_downloader.url_map.items():
+            self._assert(
+                isinstance(pattern, str) and isinstance(repl, str),
+                "config.utils.file_downloader.url_map should be a dict[str, str].",
+            )
+
+        self._assert(
+            isinstance(self.utils.journal_abbreviations, list),
+            "config.utils.journal_abbreviations should be a list.",
+        )
+        for abbrev in self.utils.journal_abbreviations:
+            self._assert(
+                isinstance(abbrev, tuple),
+                "config.utils.journal_abbreviations should be a list of tuples.",
+            )
+
     @staticmethod
     def _assert(expression: bool, error: str) -> None:
         """Asserts the expression is True.
@@ -501,6 +564,7 @@ class Config(dict):
 
     def defaults(self) -> None:
         """Resets the configuration to the default settings."""
+        # pylint: disable=consider-using-dict-items,consider-iterating-dictionary
         for section in self.DEFAULTS.keys():
             self[section].update(**self.DEFAULTS[section])
 
